//
//  ContentView.swift
//  EPUBReaderReadium
//
//  Created by Narek on 29.05.24.
//

import R2Navigator
import R2Shared
import R2Streamer
import ReadiumAdapterGCDWebServer
import SwiftUI
import WebKit

struct SwiftUIViewController: UIViewControllerRepresentable {
    @Binding var navigator: EPUBNavigatorViewController?
    @Binding var locator: Locator?
    let fileURL: URL
    func makeUIViewController(context: Context) -> some UIViewController {
        let vc = ViewController()
        vc.delegate = context.coordinator
        do {
            try vc.parseEPUB(at: fileURL)
        } catch {
            print("Error: \(error.localizedDescription)")
        }
        return vc
    }

    func updateUIViewController(_ uiViewController: UIViewControllerType, context: Context) {}

    func makeCoordinator() -> Coordinator {
        Coordinator(navigator: $navigator, locator: $locator)
    }

    class Coordinator: NSObject, ViewControllerDelegate {
        @Binding var navigator: EPUBNavigatorViewController?
        @Binding var locator: Locator?

        init(navigator: Binding<EPUBNavigatorViewController?>, locator: Binding<Locator?>) {
            self._navigator = navigator
            self._locator = locator
        }

        func epubNavigatorDidChanged(_ epubNavigator: R2Navigator.EPUBNavigatorViewController) {
            navigator = epubNavigator
        }

        func locatorDidChanged(_ locator: R2Shared.Locator) {
            self.locator = locator
        }
    }
}

protocol ViewControllerDelegate: AnyObject {
    func epubNavigatorDidChanged(_ epubNavigator: EPUBNavigatorViewController)
    func locatorDidChanged(_ locator: Locator)
}

class ViewController: UIViewController {
    var currentPage: Int = 0
    weak var delegate: ViewControllerDelegate?
    var epubNavigator: EPUBNavigatorViewController! {
        didSet {
            delegate?.epubNavigatorDidChanged(epubNavigator)
        }
    }

    init() {
        super.init(nibName: nil, bundle: nil)
    }

    @available(*, unavailable)
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    func parseEPUB(at fileURL: URL) throws {
        let asset = FileAsset(url: fileURL)
        let streamer = Streamer()
        streamer.open(asset: asset, allowUserInteraction: false) { result in
            switch result {
            case .success(let publication):
                if publication.conforms(to: .epub) {
                    do {
                        let resources = Bundle.main.bundleURL
                        let navigator = try EPUBNavigatorViewController(
                            publication: publication,
                            initialLocation: nil,
                            config: .init(
                                fontFamilyDeclarations: [
                                    CSSFontFamilyDeclaration(
                                        fontFamily: .hanalei,
                                        fontFaces: [
                                            CSSFontFace(
                                                file: resources
//                                                    .appendingPathComponent(
//                                                "DMSans-VariableFont_opsz,wght.ttf"),
//                                                    .appendingPathComponent("hanalei1.ttf"),
                                                    .appendingPathComponent(
                                                "hanalei.ttf"),
                                                style: .normal, weight: .variable(200 ... 900)
                                            )
                                        ]
                                    ).eraseToAnyHTMLFontFamilyDeclaration(),
                                ]

                            ),
                            httpServer: GCDHTTPServer.shared
                        )

                        let preferences = EPUBPreferences(
                            fontFamily: .hanalei,
                            fontSize: 1,
                            publisherStyles: false,
                            spread: .always
                        )
                        navigator.submitPreferences(preferences)
                        navigator.delegate = self
                        self.addChild(navigator)
                        navigator.view.frame = self.view.bounds
                        navigator.view.autoresizingMask = [.flexibleWidth, .flexibleHeight]
                        self.view.addSubview(navigator.view)
                        navigator.didMove(toParent: self)
                        self.epubNavigator = navigator
                    } catch {
                        print("Error: \(error.localizedDescription)")
                    }
                }
            case .failure(let error):
                print("Error: \(error)")
            case .cancelled:
                break
            }
        }
    }
}

extension ViewController: EPUBNavigatorDelegate {
    func navigator(_ navigator: any R2Navigator.Navigator, presentError error: R2Navigator.NavigatorError) {}

    func navigator(_ navigator: Navigator, locationDidChange locator: Locator) {
        delegate?.locatorDidChanged(locator)
    }

    func navigator(_ navigator: R2Navigator.EPUBNavigatorViewController, setupUserScripts userContentController: WKUserContentController) {}

    func middleTapHandler() {}

    func willExitPublication(documentIndex: Int, progression: Double?) {
        print("willExitPublication documentIndex: \(documentIndex)")
    }

    func didChangedDocumentPage(currentDocumentIndex: Int) {
        print("didChangedDocumentPage CurrentDocumentIndex: \(currentDocumentIndex)")
    }

    func didNavigateViaInternalLinkTap(to documentIndex: Int) {
        print("DidNavigateViaInternalLinkTap: \(documentIndex)")
    }

    func presentError(_ error: R2Navigator.NavigatorError) {
        print("presentError: \(error.localizedDescription)")
    }
}

struct ContentView: View {
    @State var navigator: EPUBNavigatorViewController?
    @State var locator: Locator?
    @State var fontSize: CGFloat = 1
    @State var currentLocation: Locator?

    var body: some View {
        ZStack {
            Color.white.ignoresSafeArea()
            VStack(spacing: 0) {
                if let url = Bundle.main.url(forResource: "test2", withExtension: "epub") {
                    SwiftUIViewController(
                        navigator: $navigator,
                        locator: $locator,
                        fileURL: url
                    )
                }
                Spacer().frame(height: 10)
                VStack(spacing: 0) {
                    HStack(spacing: 0) {
                        if let position = locator?.locations.position,
                           let total = navigator?.publication.positions.count
                        {
                            Text("\(position) / \(total)")
                                .foregroundStyle(Color.black)
                                .padding()
                                .background(Color.blue.opacity(0.2))
                                .clipShape(RoundedRectangle(cornerRadius: 16))

                            Spacer().frame(width: 4)
                        }
                        Button {
                            // Plus action
                            if let navigator {
                                if fontSize < 2 {
                                    fontSize += 0.25
                                    let preferences = EPUBPreferences(
                                        fontSize: fontSize,
                                        publisherStyles: false,
                                        spread: .always
                                    )
                                    let editor = navigator.editor(of: preferences)
                                    editor.fontFamily.set(.hanalei)
                                    navigator.submitPreferences(editor.preferences)
                                }
                            }
                        } label: {
                            Text("+")
                                .foregroundStyle(Color.black)
                                .padding()
                                .background(Color.blue.opacity(0.2))
                                .clipShape(RoundedRectangle(cornerRadius: 16))
                        }
                        Spacer().frame(width: 4)
                        Button {
                            // Minus Action
                            if let navigator {
                                if fontSize > 1 {
                                    fontSize -= 0.25
                                    let preferences = EPUBPreferences(
                                        fontSize: fontSize,
                                        publisherStyles: false,
                                        spread: .always
                                    )
                                    let editor = navigator.editor(of: preferences)
                                    editor.fontFamily.set(.hanalei)
                                    navigator.submitPreferences(editor.preferences)
                                }
                            }
                        } label: {
                            Text("-")
                                .foregroundStyle(Color.black)
                                .padding()
                                .background(Color.blue.opacity(0.2))
                                .clipShape(RoundedRectangle(cornerRadius: 16))
                        }
                        Spacer().frame(width: 4)
                        Button {
                            // Locator Action
                            if let currentLocation = navigator?.currentLocation {
                                self.currentLocation = currentLocation
                            }
                        } label: {
                            Text("✅")
                                .foregroundStyle(Color.black)
                                .padding()
                                .background(Color.blue.opacity(0.2))
                                .clipShape(RoundedRectangle(cornerRadius: 16))
                        }
                        Spacer().frame(width: 4)
                        Button {
                            // Locator Action
                            if let navigator, let currentLocation {
                                navigator.go(to: currentLocation)
                            }
                        } label: {
                            Text("➡")
                                .foregroundStyle(Color.black)
                                .padding()
                                .background(Color.blue.opacity(0.2))
                                .clipShape(RoundedRectangle(cornerRadius: 16))
                        }
                    } //: HSTACK
                } //: VSTACK
                Spacer().frame(height: 30)
            }
            .ignoresSafeArea(.all)
        }
    }
}

public extension FontFamily {
    static let hanalei: FontFamily = "hanalei_fill" // "Montserrat"
    static let dmSans: FontFamily = "DMSans"
}

#Preview {
    ContentView()
}
